// Copyright © 2022-2025 Gerald B. Cox
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package main

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"log/slog"
	"net/http"
	"os"
	"os/signal"
	"os/user"
	"strings"
	"syscall"

	"github.com/godbus/dbus/v5"
)

type Item struct {
	Type         string            `json:"Type"`
	Name         string            `json:"Name"`
	Album        string            `json:"Album"`
	AlbumArtist  string            `json:"AlbumArtist"`
	Artists      []string          `json:"Artists"`
	ExternalUrls []string          `json:"ExternalUrls"`
	ProviderIds  map[string]string `json:"ProviderIds"`
}

type Request struct {
	Event string `json:"Event"`
	User  struct {
		Name string `json:"Name"`
	} `json:"User"`
	Item Item `json:"Item"`
}

func getCurrentUser() string {
	u, err := user.Current()
	if err != nil {
		slog.Warn("Could not determine current user", "error", err)
		return "unknown"
	}
	return u.Username
}

var (
	AppVersionFull  = "dev"
	CommitFull      = "none"
	ShortCommitFull = "none"
	webhookURL      = os.Getenv("MPRIS_WEBHOOK_URL")
	defaultUser     = getCurrentUser()
	allowlistEnv    = os.Getenv("MPRIS_ALLOWLIST")
	allowlist       = make(map[string]bool)
)

func initAllowlist() {
	for _, name := range strings.Split(allowlistEnv, ",") {
		trimmed := strings.TrimSpace(name)
		if trimmed != "" {
			allowlist[trimmed] = true
		}
	}
	if len(allowlist) == 0 {
		slog.Error("MPRIS_ALLOWLIST is empty or invalid. Please set it to a comma-separated list of allowed players (e.g. 'qmmp,elisa').")
		os.Exit(1)
	}

	slog.Info("Configured player allowlist", "players", allowlist)
}

func getPlayerName(service string) string {
	parts := strings.Split(service, ".")
	if len(parts) >= 1 {
		return parts[len(parts)-1]
	}
	return service
}

func resolveWellKnownName(conn *dbus.Conn, uniqueName string) (string, error) {
	var names []string
	listCall := conn.BusObject().Call("org.freedesktop.DBus.ListNames", 0)

	if listCall.Err != nil {
		return "", fmt.Errorf("failed to call ListNames: %w", listCall.Err)
	}
	if err := listCall.Store(&names); err != nil {
		return "", fmt.Errorf("failed to parse ListNames result: %w", err)
	}
	for _, name := range names {
		if strings.HasPrefix(name, "org.mpris.MediaPlayer2.") {
			var owner string
			err := conn.BusObject().Call("org.freedesktop.DBus.GetNameOwner", 0, name).Store(&owner)
			if err == nil && owner == uniqueName {
				return name, nil
			}
		}
	}
	return "", fmt.Errorf("no well-known name found for %s", uniqueName)
}

func addMatchForPlayer(conn *dbus.Conn, name string) error {
	match := fmt.Sprintf("type='signal',sender='%s',interface='org.freedesktop.DBus.Properties',member='PropertiesChanged'", name)
	call := conn.BusObject().Call("org.freedesktop.DBus.AddMatch", 0, match)
	if call.Err != nil {
		return fmt.Errorf("failed to add match for %s: %w", name, call.Err)
	}
	slog.Info("Registered match rule for player", "name", name)
	return nil
}

func watchNameOwnerChanges(conn *dbus.Conn, allowlist map[string]bool, registered map[string]bool) {
	match := "type='signal',interface='org.freedesktop.DBus',member='NameOwnerChanged'"
	conn.BusObject().Call("org.freedesktop.DBus.AddMatch", 0, match)

	sigs := make(chan *dbus.Signal, 16)
	conn.Signal(sigs)

	go func() {

		for sig := range sigs {
			if len(sig.Body) != 3 {
				continue
			}
			name, ok := sig.Body[0].(string)
			if !ok || !strings.HasPrefix(name, "org.mpris.MediaPlayer2.") {
				continue
			}
			player := getPlayerName(name)
			if allowlist[player] && !registered[name] {
				if err := addMatchForPlayer(conn, name); err == nil {
					registered[name] = true
				}
			}
		}
	}()
}

func unwrapVariant(val interface{}) interface{} {
	switch v := val.(type) {
	case dbus.Variant:
		return unwrapVariant(v.Value())
	case map[string]dbus.Variant:
		m := make(map[string]interface{})
		for k, vv := range v {
			m[k] = unwrapVariant(vv)
		}
		return m
	default:
		return v
	}
}

func extractItem(meta map[string]interface{}) Item {
	item := Item{
		Type:         "Audio",
		ExternalUrls: []string{},
		ProviderIds:  map[string]string{},
	}
	if v, ok := meta["xesam:title"].(string); ok {
		item.Name = v
	}
	if v, ok := meta["xesam:album"].(string); ok {
		item.Album = v
	}
	if v, ok := meta["xesam:albumArtist"].([]string); ok && len(v) > 0 {
		item.AlbumArtist = v[0]
	}
	if v, ok := meta["xesam:artist"].([]string); ok {
		item.Artists = v
	}
	return item
}

func postPayload(event, service string, item Item) {
	req := Request{Event: event, Item: item}
	req.User.Name = defaultUser

	data, err := json.Marshal(req)
	if err != nil {
		slog.Error("Failed to marshal webhook payload", "error", err)
		return
	}

	slog.Debug("Prepared MPRIS webhook payload", "event", event, "track", item.Name)

	resp, err := http.Post(webhookURL, "application/json", strings.NewReader(string(data)))
	if err != nil {
		slog.Warn("Webhook POST failed", "error", err)
		return
	}
	defer resp.Body.Close()
	slog.Info("Webhook sent", "event", event, "track", item.Name, "status", resp.StatusCode)
}

func main() {
	versionFlag := flag.Bool("version", false, "Print version information and exit")
	flag.Parse()

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	quit := make(chan os.Signal, 1)
	signal.Notify(quit, syscall.SIGINT, syscall.SIGTERM)

	go func() {
		sig := <-quit
		slog.Info("Received shutdown signal", "signal", sig.String())
		cancel()
	}()

	if *versionFlag {
		fmt.Printf("MPRIS Webhook Version: %s\n", AppVersionFull)
		fmt.Printf("Git Commit: %s (short: %s)\n", CommitFull, ShortCommitFull)
		os.Exit(0)
	}

	var logLevel slog.Level
	switch strings.ToLower(os.Getenv("MPRIS_WEBHOOK_LOG_LEVEL")) {
	case "debug":
		logLevel = slog.LevelDebug
	case "info":
		logLevel = slog.LevelInfo
	case "warn":
		logLevel = slog.LevelWarn
	case "error":
		logLevel = slog.LevelError
	default:
		logLevel = slog.LevelInfo
	}

	slog.SetDefault(slog.New(slog.NewTextHandler(os.Stderr, &slog.HandlerOptions{
		Level: logLevel,
	})))
	slog.Info("MPRIS logging level set", "log_level", logLevel.String())
	slog.Info("Starting MPRIS Webhook", "version", AppVersionFull, "commit", CommitFull, "shortCommit", ShortCommitFull)

	if webhookURL == "" {
		slog.Error("Missing required environment variable: MPRIS_WEBHOOK_URL")
		os.Exit(1)
	}

	initAllowlist()

	slog.Info("Configured webhook endpoint", "url", webhookURL)
	slog.Info("Default scrobble user", "user", defaultUser)

	conn, err := dbus.ConnectSessionBus()
	if err != nil {
		slog.Error("Failed to connect to D-Bus session", "error", err)
		os.Exit(1)
	}
	defer conn.Close()
	slog.Info("Connected to D-Bus session")

	registered := make(map[string]bool)

	var names []string
	listCall := conn.BusObject().Call("org.freedesktop.DBus.ListNames", 0)
	if listCall.Err != nil {
		slog.Error("Failed to list D-Bus names", "error", listCall.Err)
		os.Exit(1)
	}
	if err := listCall.Store(&names); err != nil {
		slog.Error("Failed to parse D-Bus name list", "error", err)
		os.Exit(1)
	}

	for _, name := range names {
		if strings.HasPrefix(name, "org.mpris.MediaPlayer2.") {
			player := getPlayerName(name)
			if allowlist[player] {
				if err := addMatchForPlayer(conn, name); err == nil {
					registered[name] = true
				}
			}
		}
	}

	watchNameOwnerChanges(conn, allowlist, registered)

	sigs := make(chan *dbus.Signal, 16)
	conn.Signal(sigs)

	lastSeen := make(map[string]Item)

	for {
		select {
		case <-ctx.Done():
			slog.Info("Shutting down MPRIS webhook listener gracefully")
			return
		case sig := <-sigs:
			if sig == nil || len(sig.Body) < 2 {
				continue
			}

			if sig.Sender == "org.freedesktop.DBus" {
				continue
			}

			wellKnown, err := resolveWellKnownName(conn, sig.Sender)
			if err != nil {
				slog.Debug("Could not resolve well-known name", "sender", sig.Sender, "error", err)
				continue
			}
			playerName := getPlayerName(wellKnown)

			if !allowlist[playerName] {
				slog.Debug("Ignoring signal from non-allowlisted player", "service", sig.Sender, "resolved", wellKnown, "player", playerName)
				continue
			}

			iface, ok := sig.Body[0].(string)
			if !ok || iface != "org.mpris.MediaPlayer2.Player" {
				continue
			}

			props := sig.Body[1].(map[string]dbus.Variant)
			service := sig.Sender

			if statVar, ok := props["PlaybackStatus"]; ok {
				status := statVar.Value().(string)
				slog.Debug("PlaybackStatus update", "service", service, "status", status)

				if status == "Stopped" {
					if last, ok := lastSeen[service]; ok {
						postPayload("playback.stop", service, last)
						delete(lastSeen, service)
					}
					continue
				}
			}

			if metaVar, ok := props["Metadata"]; ok {
				meta := unwrapVariant(metaVar).(map[string]interface{})
				item := extractItem(meta)

				if item.Name == "" {
					continue
				}
				last, exists := lastSeen[service]
				if !exists || last.Name != item.Name {
					if exists {
						postPayload("playback.stop", service, last)
					}
					postPayload("playback.start", service, item)
					lastSeen[service] = item
				}
			}
		}
	}
}
